import json
from typing import List, Optional, Union

from duowen_agent.agents.base import BaseToolResult
from duowen_agent.error import ToolError
from duowen_agent.tools.base import BaseTool


class ToolManager:
    """ToolManager helps Agent to manage tools"""

    def __init__(self, tools: List[BaseTool], filter_function_list: List[str] = None):
        self.tools: List[BaseTool] = tools
        self.filter_function_list = filter_function_list

    def get_tool(self, tool_name: str) -> Optional[BaseTool]:
        """Find specified tool by tool name.
        Args:
            tool_name(str): The name of the tool.

        Returns:
            Optional[BaseTool]: The specified tool or None if not found.
        """
        return next((tool for tool in self.tools if tool.name == tool_name), None)

    def run_tool(
        self, tool_name: str, parameters: Union[str, dict]
    ) -> BaseToolResult | str | int | float:
        """Run tool by input tool name and data inputs

        Args:
            tool_name(str): The name of the tool.
            parameters(Union[str, dict]): The parameters for the tool.

        Returns:
            str: The result of the tool.
            Any: 结构化数据用于页面展示
        """
        tool = self.get_tool(tool_name)

        if tool is None:
            return (
                f"{tool_name} has not been provided yet, please use the provided tool."
            )

        if isinstance(parameters, dict):
            data = tool.run(**parameters)
        else:
            data = tool.run(parameters)

        if data:
            return data
        else:
            raise ToolError(f"Tool {tool_name} has no return value.")

    @property
    def tool_names(self) -> str:
        """Get all tool names."""
        tool_names = ""
        for tool in self.tools:
            tool_names += f"{tool.name}, "
        return tool_names[:-2]

    @property
    def tool_classifiers(self) -> dict[str, str]:
        _classifiers = {}
        if self.filter_function_list:
            for tool in self.tools:
                if tool.name in self.filter_function_list:
                    _classifiers[tool.name] = tool.description
        else:
            for tool in self.tools:
                _classifiers[tool.name] = tool.description
        return _classifiers

    @property
    def tool_descriptions(self) -> str:
        """Get all tool descriptions, including the schema if available."""
        tool_descriptions = ""
        if self.filter_function_list:
            for tool in self.tools:
                if tool.name in self.filter_function_list:
                    tool_descriptions += (
                        json.dumps(
                            tool.to_schema(),
                            ensure_ascii=False,
                        )
                        + "\n"
                    )
        else:
            for tool in self.tools:
                tool_descriptions += (
                    json.dumps(
                        tool.to_schema(),
                        ensure_ascii=False,
                    )
                    + "\n"
                )
        return tool_descriptions



import re
import threading
import time
from collections import defaultdict, Counter
from typing import List

from duowen_agent.llm import OpenAIChat, MessagesSet
from duowen_agent.rag.graph.base import BaseLLM
from duowen_agent.utils.concurrency import concurrent_execute

from . import BaseKVStorage
from .prompt import (
    DEFAULT_TUPLE_DELIMITER,
    DEFAULT_RECORD_DELIMITER,
    DEFAULT_COMPLETION_DELIMITER,
    GRAPH_EXTRACTION_PROMPT,
    CONTINUE_PROMPT,
    LOOP_PROMPT,
    SUMMARIZE_DESCRIPTIONS_PROMPT,
    GRAPH_FIELD_SEP,
)
from .utils import (
    print_call_back,
    is_interrupt,
    clean_str,
    is_float_regex,
    flat_uniq_list,
    truncate,
)


class Extractor(BaseLLM):
    def __init__(
        self,
        llm_instance: OpenAIChat,
        language: str = "Chinese",
        entity_types: List[str] = None,
        max_gleanings: int = 2,
        call_back_func: callable = print_call_back,
        interrupt_func: callable = is_interrupt,
        retry_cnt: int = 3,
        retry_sleep: int = 1,
        concurrent_num: int = 1,
        llm_cache: BaseKVStorage = None,
    ):
        super().__init__(
            llm_instance,
            call_back_func,
            interrupt_func,
            retry_cnt,
            retry_sleep,
            concurrent_num,
            llm_cache,
        )

        self._language = language
        self._entity_types = entity_types or [
            "organization",
            "person",
            "geo",
            "event",
            "category",
        ]
        self._call_back_func = call_back_func
        self._interrupt_func = interrupt_func
        self.max_gleanings = max_gleanings
        self.entity_extract_prompt = GRAPH_EXTRACTION_PROMPT
        self.context_base = dict(
            tuple_delimiter=DEFAULT_TUPLE_DELIMITER,
            record_delimiter=DEFAULT_RECORD_DELIMITER,
            completion_delimiter=DEFAULT_COMPLETION_DELIMITER,
            entity_types=", ".join(self._entity_types),
        )
        self.continue_prompt = CONTINUE_PROMPT
        self.if_loop_prompt = LOOP_PROMPT
        self.total_cnt = 0
        self.current_cnt = 0
        self._lock = threading.Lock()

    def extract(self, chunks: dict[str, str]) -> tuple[list, list]:
        _params_lst = [{"chunk_key": k, "chunk_content": v} for k, v in chunks.items()]
        self.total_cnt = len(_params_lst)
        self.current_cnt = 0
        out_results = concurrent_execute(
            self._process_single_content,
            _params_lst,
            work_num=self.concurrent_num,
        )

        maybe_nodes = defaultdict(list)
        maybe_edges = defaultdict(list)
        for i in out_results:
            m_nodes, m_edges = i
            for k, v in m_nodes.items():
                maybe_nodes[k].extend(v)
            for k, v in m_edges.items():
                maybe_edges[tuple(sorted(k))].extend(v)

        self._call_back_func(
            msg=f"实体和关系提取完成, {len(maybe_nodes)} 节点, {len(maybe_edges)} 关系."
        )

        self._call_back_func("开始实体合并")
        all_entities_data = []

        concurrent_execute(
            self._merge_nodes,
            [
                {
                    "entity_name": en_nm,
                    "entities": ents,
                    "all_entities_data": all_entities_data,
                }
                for en_nm, ents in maybe_nodes.items()
            ],
            work_num=self.concurrent_num,
        )

        self._call_back_func("实体合并完成")
        self._call_back_func("开始关系合并")

        all_relationships_data = []
        concurrent_execute(
            self._merge_edges,
            [
                {
                    "src_id": src,
                    "tgt_id": tgt,
                    "edges_data": rels,
                    "all_relationships_data": all_relationships_data,
                }
                for (src, tgt), rels in maybe_edges.items()
            ],
            work_num=self.concurrent_num,
        )

        self._call_back_func("关系合并完成")

        if not len(all_entities_data) and not len(all_relationships_data):
            self._call_back_func("没有提取任何实体和关系，可能是您的语言模型异常")

        if not len(all_entities_data):
            self._call_back_func("没有抽取到任何实体")
        if not len(all_relationships_data):
            self._call_back_func("没有抽取任何关系")

        return all_entities_data, all_relationships_data

    def _process_single_content(
        self, chunk_key: str, chunk_content: str
    ) -> tuple[dict, dict]:
        _prompt = MessagesSet()
        _prompt.add_system(
            self.entity_extract_prompt.format(
                **{**self.context_base, "input_text": chunk_content}
            )
        )
        _prompt.add_user("Output:")
        response = self._chat(_prompt, temperature=0.3)
        results = response or ""
        _prompt.add_assistant(response)
        for i in range(self.max_gleanings):
            _prompt.add_user(self.continue_prompt)
            response = self._chat(_prompt, temperature=0.3)
            results += response or ""

            if i >= self.max_gleanings - 1:
                break

            _prompt.add_assistant(response)
            _prompt.add_user(self.if_loop_prompt)
            # print(_prompt.get_format_messages())
            continuation = self._chat(_prompt, temperature=0.8)
            # print(continuation)
            if continuation.strip() != "Y":
                break
            _prompt.add_assistant("Y")

        records = self._split_string_by_multi_markers(
            results,
            [
                DEFAULT_RECORD_DELIMITER,
                DEFAULT_COMPLETION_DELIMITER,
            ],
        )

        rcds = []
        for record in records:
            record = re.search(r"\((.*)\)", record)
            if record is None:
                continue
            rcds.append(record.group(1))
        records = rcds

        maybe_nodes, maybe_edges = self._entities_and_relations(
            chunk_key, records, DEFAULT_TUPLE_DELIMITER
        )
        with self._lock:
            self.current_cnt += 1

            self._call_back_func(
                msg=f"实体抽取({'、'.join([i for i in maybe_nodes])[:10]}...)，已处理 {self.current_cnt}/{self.total_cnt} 个 chunk"
            )

        return maybe_nodes, maybe_edges

    def _handle_entity_relation_summary(
        self, entity_or_relation_name: str, description: str
    ) -> str:
        summary_max_tokens = 512
        use_description = truncate(description, summary_max_tokens)
        description_list = (use_description.split(GRAPH_FIELD_SEP),)
        if len(description_list) <= 12:
            return use_description
        prompt_template = SUMMARIZE_DESCRIPTIONS_PROMPT
        context_base = dict(
            entity_name=entity_or_relation_name,
            description_list=description_list,
            language=self._language,
        )
        use_prompt = prompt_template.format(**context_base)
        _prompt = MessagesSet()

        _prompt.add_system(use_prompt)
        _prompt.add_user("Output: ")

        self._call_back_func(f"触发摘要: {entity_or_relation_name}")
        summary = self._chat(_prompt, temperature=0.8)
        return summary

    def _merge_nodes(self, entity_name: str, entities: list[dict], all_entities_data):
        if not entities:
            return
        entity_type = sorted(
            Counter([dp["entity_type"] for dp in entities]).items(),
            key=lambda x: x[1],
            reverse=True,
        )[0][0]
        description = GRAPH_FIELD_SEP.join(
            sorted(set([dp["description"] for dp in entities]))
        )
        already_source_ids = flat_uniq_list(entities, "source_id")
        description = self._handle_entity_relation_summary(entity_name, description)
        node_data = dict(
            entity_type=entity_type,
            description=description,
            source_id=already_source_ids,
        )
        node_data["entity_name"] = entity_name
        with self._lock:
            all_entities_data.append(node_data)

    def _merge_edges(
        self,
        src_id: str,
        tgt_id: str,
        edges_data: list[dict],
        all_relationships_data=None,
    ):
        if not edges_data:
            return
        weight = sum([edge["weight"] for edge in edges_data])
        description = GRAPH_FIELD_SEP.join(
            sorted(set([edge["description"] for edge in edges_data]))
        )
        description = self._handle_entity_relation_summary(
            f"{src_id} -> {tgt_id}", description
        )
        source_id = flat_uniq_list(edges_data, "source_id")
        edge_data = dict(
            src_id=src_id,
            tgt_id=tgt_id,
            description=description,
            weight=weight,
            source_id=source_id,
        )
        with self._lock:
            all_relationships_data.append(edge_data)

    @staticmethod
    def _handle_single_relationship_extraction(
        record_attributes: list[str], chunk_key: str
    ):
        if len(record_attributes) < 5 or record_attributes[0] != '"relationship"':
            return None
        # add this record as edge
        source = clean_str(record_attributes[1].upper())
        target = clean_str(record_attributes[2].upper())
        edge_description = clean_str(record_attributes[3])

        edge_source_id = chunk_key
        weight = (
            float(record_attributes[-1])
            if is_float_regex(record_attributes[-1])
            else 1.0
        )
        pair = sorted([source.upper(), target.upper()])
        return dict(
            src_id=pair[0],
            tgt_id=pair[1],
            weight=weight,
            description=edge_description,
            source_id=edge_source_id,
            metadata={"created_at": time.time()},
        )

    @staticmethod
    def _handle_single_entity_extraction(
        record_attributes: list[str],
        chunk_key: str,
    ):
        if len(record_attributes) < 4 or record_attributes[0] != '"entity"':
            return None
        # add this record as a node in the G
        entity_name = clean_str(record_attributes[1].upper())
        if not entity_name.strip():
            return None
        entity_type = clean_str(record_attributes[2].upper())
        entity_description = clean_str(record_attributes[3])
        entity_source_id = chunk_key
        return dict(
            entity_name=entity_name.upper(),
            entity_type=entity_type.upper(),
            description=entity_description,
            source_id=entity_source_id,
        )

    def _entities_and_relations(
        self, chunk_key: str, records: list, tuple_delimiter: str
    ):
        maybe_nodes = defaultdict(list)
        maybe_edges = defaultdict(list)
        ent_types = [t.lower() for t in self._entity_types]
        for record in records:
            record_attributes = self._split_string_by_multi_markers(
                record, [tuple_delimiter]
            )

            if_entities = self._handle_single_entity_extraction(
                record_attributes, chunk_key
            )
            if (
                if_entities is not None
                and if_entities.get("entity_type", "unknown").lower() in ent_types
            ):
                maybe_nodes[if_entities["entity_name"]].append(if_entities)
                continue

            if_relation = self._handle_single_relationship_extraction(
                record_attributes, chunk_key
            )
            if if_relation is not None:
                maybe_edges[(if_relation["src_id"], if_relation["tgt_id"])].append(
                    if_relation
                )
        return dict(maybe_nodes), dict(maybe_edges)

    @staticmethod
    def _split_string_by_multi_markers(content: str, markers: list[str]) -> list[str]:
        """Split a string by multiple markers"""
        if not markers:
            return [content]
        results = re.split("|".join(re.escape(marker) for marker in markers), content)
        return [r.strip() for r in results if r.strip()]



from dataclasses import dataclass
from typing import Any, Generic, TypeVar

from duowen_agent.mcp.session.base_session import BaseSession
from duowen_agent.mcp.types import LATEST_PROTOCOL_VERSION, RequestId, RequestParams

SUPPORTED_PROTOCOL_VERSIONS: list[str] = ["2024-11-05", LATEST_PROTOCOL_VERSION]


SessionT = TypeVar("SessionT", bound=BaseSession[Any, Any, Any, Any, Any])
LifespanContextT = TypeVar("LifespanContextT")


@dataclass
class RequestContext(Generic[SessionT, LifespanContextT]):
    request_id: RequestId
    meta: RequestParams.Meta | None
    session: SessionT
    lifespan_context: LifespanContextT



import json
import logging
from datetime import date
from os import getenv
from typing import Optional

import requests
from duowen_agent.error import ToolError
from duowen_agent.tools.base import BaseTool
from pydantic import BaseModel, Field, field_validator

from .entity import ToolSearchResult, ToolSearchResultDetails


class BochaParameters(BaseModel):
    """检索查询参数模型"""

    query: str = Field(..., min_length=1, description="用户的搜索词（必填）")
    freshness: Optional[str] = Field(
        default="noLimit",
        description='搜索指定时间范围内的网页。\n可填值：\n- oneDay，一天内\n- oneWeek，一周内\n- oneMonth，一个月内\n- oneYear，一年内\n- noLimit，不限（默认）\n- YYYY-MM-DD..YYYY-MM-DD，搜索日期范围，例如："2025-01-01..2025-04-06"\n- YYYY-MM-DD，搜索指定日期，例如："2025-04-06"',
    )
    summary: Optional[bool] = Field(default=True, description="是否显示文本摘要")
    count: Optional[int] = Field(
        default=10, ge=1, le=50, description="返回结果数量（1-50）"
    )
    page: Optional[int] = Field(default=1, ge=1, description="页码（≥1）")

    @field_validator("freshness")
    def validate_freshness(cls, v):
        """验证时间范围参数"""
        allowed_values = {"oneDay", "oneWeek", "oneMonth", "oneYear", "noLimit"}

        if v in allowed_values:
            return v

        # 验证日期范围格式
        if ".." in v:
            dates = v.split("..")
            if len(dates) != 2:
                raise ToolError("日期范围格式错误，使用 'YYYY-MM-DD..YYYY-MM-DD'")
            start_date, end_date = dates

            try:
                # 验证日期合法性
                date.fromisoformat(start_date)
                date.fromisoformat(end_date)
            except ValueError:
                raise ToolError("非法日期格式，使用 YYYY-MM-DD")

            return v

        # 验证单日期格式
        try:
            date.fromisoformat(v)
            return v
        except ValueError:
            raise ValueError("非法日期格式，使用 YYYY-MM-DD 或预设值")


class Bocha(BaseTool):

    name: str = "bocha-search-api"
    description: str = (
        "专业网络搜索引擎，支持深度内容抓取和定制过滤。"
        "优势：精准的来源控制、时效性筛选、结构化数据输出。"
        "适用场景：新闻追踪、学术研究、市场分析等需要最新网络信息的任务。"
    )
    parameters = BochaParameters

    def __init__(
        self,
        base_url: str = None,
        api_key: str = None,
        token_limit: int = 8000,
        max_results: int = 5,
    ):
        super().__init__()
        self.base_url = base_url or "https://api.bochaai.com/v1/web-search"
        self.api_key = api_key or getenv("BOCHA_API_KEY")
        self.token_limit = token_limit
        self.page = max_results

    def search(self, query: str, **kwargs) -> ToolSearchResult:

        _param = {"query": query}
        _param.update(kwargs)

        _bocha_param = BochaParameters(**_param)
        try:

            payload = json.dumps(_bocha_param.model_dump(), ensure_ascii=False)
            headers = {
                "Authorization": f"Bearer {self.api_key}",
                "Content-Type": "application/json",
            }

            response = requests.request(
                "POST", self.base_url, headers=headers, data=payload
            )

            if response.status_code == 200:

                data1 = []
                _token_limit = self.token_limit
                for i in response.json()["data"]["webPages"]["value"][: self.page]:
                    _res = ToolSearchResultDetails(
                        url=i["url"],
                        title=i["name"],
                        content=i.get("summary", None) or i.get("snippet", None),
                        date_published=i["datePublished"],
                        site_name=i["siteName"],
                        site_icon=i["siteIcon"],
                    )
                    if _res.chat_token <= _token_limit:
                        data1.append(_res)
                        _token_limit -= _res.chat_token
                    else:
                        break

                return ToolSearchResult(result=data1)
        except Exception as e:
            logging.exception(e)
            raise ToolError(str(e))

    def _run(
        self,
        query: str,
        **kwargs,
    ):
        _data = self.search(query, **kwargs)
        return _data



import logging
from collections.abc import Callable
from contextlib import AbstractContextManager, ExitStack
from types import TracebackType
from typing import Any, Optional, cast
from urllib.parse import urlparse

from duowen_agent.mcp.client.sse_client import sse_client
from duowen_agent.mcp.client.streamable_client import streamablehttp_client
from duowen_agent.mcp.error import MCPAuthError, MCPConnectionError
from duowen_agent.mcp.session.client_session import ClientSession
from duowen_agent.mcp.types import Tool, OAuthTokens

logger = logging.getLogger(__name__)


class MCPClient:
    def __init__(
        self,
        server_url: str,
        authed: bool = True,
        token: Optional[OAuthTokens] = None,
        authorization_code: Optional[str] = None,
    ):
        # Initialize info
        self.client_type = "streamable"
        self.server_url = server_url

        # Authentication info
        self.authed = authed
        self.authorization_code = authorization_code
        if authed:
            self.token = token

        # Initialize session and client objects
        self._session: Optional[ClientSession] = None
        self._streams_context: Optional[AbstractContextManager[Any]] = None
        self._session_context: Optional[ClientSession] = None
        self.exit_stack = ExitStack()

        # Whether the client has been initialized
        self._initialized = False

    def __enter__(self):
        self._initialize()
        self._initialized = True
        return self

    def __exit__(
        self,
        exc_type: Optional[type],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ):
        self.cleanup()

    def _initialize(
        self,
    ):
        """Initialize the client with fallback to SSE if streamable connection fails"""
        connection_methods: dict[str, Callable[..., AbstractContextManager[Any]]] = {
            "mcp": streamablehttp_client,
            "sse": sse_client,
        }

        parsed_url = urlparse(self.server_url)
        path = parsed_url.path or ""
        method_name = path.rstrip("/").split("/")[-1] if path else ""
        if method_name in connection_methods:
            client_factory = connection_methods[method_name]
            self.connect_server(client_factory, method_name)
        else:
            try:
                logger.debug(
                    "Not supported method %s found in URL path, trying default 'mcp' method.",
                    method_name,
                )
                self.connect_server(sse_client, "sse")
            except MCPConnectionError:
                logger.debug(
                    "MCP connection failed with 'sse', falling back to 'mcp' method."
                )
                self.connect_server(streamablehttp_client, "mcp")

    def connect_server(
        self,
        client_factory: Callable[..., AbstractContextManager[Any]],
        method_name: str,
        first_try: bool = True,
    ):

        try:
            headers = (
                {
                    "Authorization": f"{self.token.token_type.capitalize()} {self.token.access_token}"
                }
                if self.authed and self.token
                else {}
            )
            self._streams_context = client_factory(url=self.server_url, headers=headers)
            if not self._streams_context:
                raise MCPConnectionError("Failed to create connection context")

            # Use exit_stack to manage context managers properly
            if method_name == "mcp":
                read_stream, write_stream, _ = self.exit_stack.enter_context(
                    self._streams_context
                )
                streams = (read_stream, write_stream)
            else:  # sse_client
                streams = self.exit_stack.enter_context(self._streams_context)

            self._session_context = ClientSession(*streams)
            self._session = self.exit_stack.enter_context(self._session_context)
            session = cast(ClientSession, self._session)
            session.initialize()
            return

        except MCPAuthError:
            if not self.authed:
                raise

            if first_try:
                return self.connect_server(client_factory, method_name, first_try=False)

        except MCPConnectionError:
            raise

    def list_tools(self) -> list[Tool]:
        """Connect to an MCP server running with SSE transport"""
        # List available tools to verify connection
        if not self._initialized or not self._session:
            raise ValueError("Session not initialized.")
        response = self._session.list_tools()
        tools = response.tools
        return tools

    def invoke_tool(self, tool_name: str, tool_args: dict):
        """Call a tool"""
        if not self._initialized or not self._session:
            raise ValueError("Session not initialized.")
        return self._session.call_tool(tool_name, tool_args)

    def cleanup(self):
        """Clean up resources"""
        try:
            # ExitStack will handle proper cleanup of all managed context managers
            self.exit_stack.close()
        except Exception as e:
            logging.exception("Error during cleanup")
            raise ValueError(f"Error during cleanup: {e}")
        finally:
            self._session = None
            self._session_context = None
            self._streams_context = None
            self._initialized = False



# Copyright (c) 2025 Bytedance Ltd. and/or its affiliates
# SPDX-License-Identifier: MIT

from typing import Literal

# Define available LLM types
LLMType = Literal["basic", "reasoning", "vision"]

# Define agent-LLM mapping
AGENT_LLM_MAP: dict[str, LLMType] = {
    "coordinator": "basic",
    "planner": "basic",
    "researcher": "basic",
    "coder": "basic",
    "reporter": "basic",
    "podcast_script_writer": "basic",
    "ppt_composer": "basic",
    "prose_writer": "basic",
    "prompt_enhancer": "basic",
}
